#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
to run on Raspberry!

UDP server for returning measured data from the spacecraft simulator testbed
listens port 6502
data return to sender port

accepts double [-1..+1] - fan control, eff. PWM

for any control packet - return 13-elem. vector
1. current control u
2. vector of 9 double elements (MPU9250)
wx, wy, wz [deg/s]
ax, ay, az [g]
mx, my, mz [mT]
3. vector of 3 double elements (QMC5883)
m2x, m2y, m2z [mT]

All vectors in instrumental coordinate system, unfiltered, unbiased 
@author: a.t.lelekov@yandex.ru
"""
import pigpio
from mpu9250 import mpu9250
import qmc5883l as qmc
import socket
import struct
import numpy as np
import logging
import sys

import serverLib as lib

#%% Main code

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(message)s',
    datefmt='%m/%d/%Y %I:%M:%S %p',
    handlers=[
        logging.FileHandler("/home/pi/serverR.log"),
        logging.StreamHandler(sys.stdout)
    ]
)

logging.info('[*] Started!')

# Fan control pins number (Broadcom)
# https://pinout.xyz/#s
cwPin = 13  # clockwise fan
ccPin = 12 # conuterclockwise fan

freq = 1000 # frequency, Hz
MEG = 1000000 # 1M, for duty cycle

pi = pigpio.pi() # pi accesses the local Pi's GPIO
if not pi.connected:
    logging.info('[!] can\'t connect with pigpio..')
    exit()
pi.hardware_PWM(ccPin, freq, 0) # stop fans...
pi.hardware_PWM(cwPin, freq, 0)

logging.info('[*] pigpio connected, starting udp server..')

imu = mpu9250()
compass = qmc.QMC5883L()
logging.info('[*] QMC5883L connected..')

bind_ip = lib.getIP()
bind_port = 6502
server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
server.bind((bind_ip,bind_port))  # Binding an address and port to a socket. 
server.settimeout(1.5)
logging.info('[*] Ready to receive AckMPU on %s:%d' % (bind_ip,bind_port))

with lib.Printer() as p:
    while True:
        sendAns = False
        try:
            ack, senderAddr = server.recvfrom(64)
            #print('[*] ack from %s:%d'%(senderAddr[0], senderAddr[1]))
            fan = struct.unpack('f', ack)[0]
            if np.isnan(fan):
                fan = 0.
            sendAns = True
            p.it()

        except struct.error:
            fan = 0.
            sendAns = True
            p.asterisk()

        except socket.timeout:
            fan = 0.
            p.asterisk()

        except KeyboardInterrupt:
            logging.warning('[!] Ctrl+C')
            break # exit from loop

        except Exception as e: # https://stackoverflow.com/a/33239954/5355749
            if hasattr(e, 'message'):
                logging.error(e.message)
            else:
                logging.error(e)
            break

        pi.hardware_PWM(ccPin, freq, int(MEG*np.clip(-fan, 0., 1.)))
        pi.hardware_PWM(cwPin, freq, int(MEG*np.clip(fan, 0., 1.)))

        if sendAns: # send data
            data = [fan]
            data.extend(imu.accel)
            data.extend(imu.gyro)
            data.extend(imu.mag)
            data.extend(compass.measure())

            # answer data in instrum. CS (also see documentation for MPU-9250, QMC) 
            msg = struct.pack('13f', *data)
            server.sendto(msg, senderAddr) # to sender address

logging.info('[*] Exiting...')
server.close()

pi.hardware_PWM(ccPin, freq, 0) # stop fans...
pi.hardware_PWM(cwPin, freq, 0)
logging.info('[*] Finish!')